package main

import (
	"encoding/json"
	"fmt"
	"log"
	"net/http"
	"os"
	"os/signal"
	"strconv"
	"strings"
	"syscall"
	"time"

	mqtt "github.com/eclipse/paho.mqtt.golang"
	"gopkg.in/ini.v1"
)

// Config is the internal singleton configuration for this program
type Config struct {
	MqttHost     string
	MqttClientID string
	InfluxHost   string
	InfluxBucket string
	InfluxOrg    string
	InfluxToken  string
	HttpBindAddr string
	Verbose      bool
}

type Node struct {
	ID           int                    `json:"id"`           // NodeID
	NodeName     string                 `json:"name"`         // Name of the measurement node
	Measurements map[string]Measurement `json:"measurements"` // Stored measurements of this node
}

func (n *Node) setMeasurement(mst Measurement) {
	n.Measurements[mst.Name] = mst
}

type Measurement struct {
	Node      int                `json:"id"`          // NodeID
	NodeName  string             `json:"name"`        // Name of the measurement node
	Timestamp int64              `json:"tsp"`         // Timestamp when the measurement was taken
	Name      string             `json:"measurement"` // Measurement name
	Values    map[string]float32 `json:"values"`      // Measurement values
	Tags      map[string]string  `json:"tags"`        // Optional tags for this measurement
}

func (mst *Measurement) json() string {
	ret := fmt.Sprintf("node %d: {", mst.Node)
	first := true
	for k, v := range mst.Values {
		if first {
			first = false
		} else {
			ret += ", "
		}
		ret += fmt.Sprintf("\"%s\":%f", k, v)
	}
	ret += "}"
	return ret
}

func (c *Config) loadIni(filename string) error {
	cfg, err := ini.Load(filename)
	if err != nil {
		return err
	}

	mqtt := cfg.Section("mqtt")
	if mqtthost := mqtt.Key("remote").String(); mqtthost != "" {
		c.MqttHost = mqtthost
	}
	if mqttclientid := mqtt.Key("clientid").String(); mqttclientid != "" {
		c.MqttClientID = mqttclientid
	}

	influx := cfg.Section("influxdb")
	if influxhost := influx.Key("remote").String(); influxhost != "" {
		c.InfluxHost = influxhost
	}
	if influxOrg := influx.Key("organization").String(); influxOrg != "" {
		c.InfluxOrg = influxOrg
	}
	if influxToken := influx.Key("token").String(); influxToken != "" {
		c.InfluxToken = influxToken
	}
	if influxdb := influx.Key("bucket").String(); influxdb != "" {
		c.InfluxBucket = influxdb
	}

	http := cfg.Section("http")
	if bindAddr := http.Key("bind_addr").String(); bindAddr != "" {
		c.HttpBindAddr = bindAddr
	}

	return nil
}

var config Config
var influx InfluxDB
var nodes map[int]Node

func receiveMeasurement(msg mqtt.Message) (Measurement, error) {
	var mst Measurement

	payload := msg.Payload()
	mst.Timestamp = time.Now().Unix()
	mst.Name = "meteo"

	// Parse node ID from topic, if present
	topic := msg.Topic()
	if len(topic) > 6 {
		var err error
		mst.Node, err = strconv.Atoi(msg.Topic()[6:])
		if err != nil {
			return mst, fmt.Errorf("invalid meteo id")
		}
	}

	// Try to parse the new data field version
	// The measurement is considered as valid, if there is values defined
	// otherwise we continue with the old data parsing
	if err := json.Unmarshal(payload, &mst); err == nil && mst.Values != nil {
		return mst, nil
	}

	// Proceed with the old data parsing
	data := make(map[string]interface{}, 0)
	if err := json.Unmarshal(msg.Payload(), &data); err != nil {
		return mst, err
	}

	// remove unnecessary fields
	delete(data, "id")
	delete(data, "name")
	delete(data, "measurement")

	// Parse fields
	mst.Values = make(map[string]float32)
	for k, v := range data {
		f, err := strconv.ParseFloat(fmt.Sprintf("%f", v), 32)
		if err != nil {
			return mst, fmt.Errorf("invalid measurement value for %s", k)
		}
		mst.Values[k] = float32(f)
	}
	return mst, nil
}

// Callback when a mqtt message has been received
func received(msg mqtt.Message) {
	mst, err := receiveMeasurement(msg)
	if err != nil {
		fmt.Fprintf(os.Stderr, "error receiving measurement: %s\n", err)
		return
	}

	err = processMeasurement(mst)
	if err != nil {
		fmt.Fprintf(os.Stderr, "%s\n", err)
		return
	}
	if config.Verbose {
		fmt.Println(mst.json())
	}
}

func processMeasurement(mst Measurement) error {
	// Update node
	if node, ok := nodes[mst.Node]; ok {
		node.setMeasurement(mst)
		nodes[mst.Node] = node
	} else {
		// Create new node
		node := Node{ID: mst.Node, NodeName: mst.NodeName}
		node.setMeasurement(mst)
		nodes[mst.Node] = node
	}

	tags := map[string]string{"node": fmt.Sprintf("%d", mst.Node)}
	for k, v := range mst.Tags {
		tags[k] = v
	}
	// Write the whole measurement at once
	fields := make(map[string]interface{}, 0)
	for k, v := range mst.Values {
		fields[k] = v
	}
	if err := influx.Write(config.InfluxOrg, config.InfluxBucket, mst.Name, tags, fields); err != nil {
		return fmt.Errorf("cannot write to influxdb: %s", err)
	}
	return nil
}

// awaits SIGINT or SIGTERM
func awaitTerminationSignal() {
	sigs := make(chan os.Signal, 1)
	done := make(chan bool, 1)
	signal.Notify(sigs, syscall.SIGINT, syscall.SIGTERM)
	go func() {
		sig := <-sigs
		fmt.Println(sig)
		done <- true
	}()
	<-done
}

func printUsage() {
	fmt.Println("meteod")
	fmt.Printf("Usage: %s [OPTIONS]\n", os.Args[0])
	fmt.Println("OPTIONS")
	fmt.Println("  -h,--help             Display this help message")
	fmt.Println("  -c,--config FILE      Load config file")
	fmt.Println("  -v, --verbose         Verbose mode")
	fmt.Println("  --mqtt MQTT           Set mqtt server")
	fmt.Println("  --mqtt-client-id ID   Set mqtt client id")
	fmt.Println("  --influx HOST         Set influxdb hostname")
	fmt.Println("  --org ORGANIZATION    Set influxdb organization")
	fmt.Println("  --bucket BUCKET       Set influxdb bucket")
	fmt.Println("  --token TOKEN         Set influxdb token")
	fmt.Println("  --http BINDADDR       Enable http server on the given address")

}

func fileExists(filename string) bool {
	if _, err := os.Stat(filename); os.IsNotExist(err) {
		return false
	}
	return true
}

func main() {
	var err error

	nodes = make(map[int]Node)

	// Default settings
	config.MqttHost = "127.0.0.1"
	config.InfluxHost = "http://127.0.0.1:8086"
	config.InfluxBucket = "meteo"
	config.InfluxToken = ""
	config.Verbose = false

	configFile := "/etc/meteo/meteod.conf"
	if fileExists(configFile) {
		if err := config.loadIni(configFile); err != nil {
			fmt.Fprintf(os.Stderr, "error loading config file %s: %s\n", configFile, err)
			os.Exit(1)
		}
	}

	args := os.Args[1:]
	for i := 0; i < len(args); i++ {
		arg := strings.TrimSpace(args[i])
		if arg == "" {
			continue
		} else if arg[0] == '-' {
			last := i >= len(args)-1
			if arg == "-h" || arg == "--help" {
				printUsage()
				return
			} else if arg == "-c" || arg == "--config" {
				if last {
					fmt.Fprintf(os.Stderr, "Missing argument: config file\n")
					os.Exit(1)
				}
				i++
				configFile = args[i]
				if err := config.loadIni(configFile); err != nil {
					fmt.Fprintf(os.Stderr, "error loading ini file %s: %s\n", configFile, err)
					os.Exit(1)
				}
			} else if arg == "-v" || arg == "--verbose" {
				config.Verbose = true
			} else if arg == "--mqtt" {
				if last {
					fmt.Fprintf(os.Stderr, "Missing argument: mqtt remote\n")
					os.Exit(1)
				}
				i++
				config.MqttHost = args[i]
			} else if arg == "--mqtt-client-id" {
				if last {
					fmt.Fprintf(os.Stderr, "Missing argument: mqtt-client-id remote\n")
					os.Exit(1)
				}
				i++
				config.MqttClientID = args[i]
			} else if arg == "--influx" {
				if last {
					fmt.Fprintf(os.Stderr, "Missing argument: influx remote\n")
					os.Exit(1)
				}
				i++
				config.InfluxHost = args[i]
			} else if arg == "--token" {
				if last {
					fmt.Fprintf(os.Stderr, "Missing argument: influx token\n")
					os.Exit(1)
				}
				i++
				config.InfluxToken = args[i]
			} else if arg == "--bucket" || arg == "--database" {
				if last {
					fmt.Fprintf(os.Stderr, "Missing argument: influx bucket\n")
					os.Exit(1)
				}
				i++
				config.InfluxBucket = args[i]
			} else if arg == "--org" || arg == "--organization" {
				if last {
					fmt.Fprintf(os.Stderr, "Missing argument: influx organization\n")
					os.Exit(1)
				}
				i++
				config.InfluxOrg = args[i]
			} else if arg == "--http" {
				if last {
					fmt.Fprintf(os.Stderr, "Missing argument: http bind address\n")
					os.Exit(1)
				}
				i++
				config.HttpBindAddr = args[i]
			} else {
				fmt.Fprintf(os.Stderr, "Invalid argument: %s\n", arg)
				os.Exit(1)

			}
		} else {
			fmt.Fprintf(os.Stderr, "Invalid argument: %s\n", arg)
			os.Exit(1)
		}
	}

	// Connect InfluxDB
	influx, err = ConnectInfluxDB(config.InfluxHost, config.InfluxToken)
	if err != nil {
		fmt.Fprintf(os.Stderr, "cannot connect to influxdb: %s\n", err)
		os.Exit(1)
	}
	if ping, err := influx.Ping(5 * time.Second); err != nil {
		fmt.Fprintf(os.Stderr, "cannot ping influxdb: %s\n", err)
		os.Exit(1)
	} else {
		if config.Verbose {
			fmt.Printf("influxdb connected (Ping: %d ms)\n", ping.Milliseconds())
		}
	}

	// Connect to mqtt server
	mqtt := MQTT{}
	mqtt.ClientID = config.MqttClientID
	mqtt.Verbose = config.Verbose
	mqtt.Callback = received
	if err := mqtt.Connect(config.MqttHost, 1883); err != nil {
		fmt.Fprintf(os.Stderr, "mqtt error: %s\n", err)
		os.Exit(1)
	}

	// Enable http server
	if config.HttpBindAddr != "" {
		http.HandleFunc("/", handlerDefault)
		http.HandleFunc("/health", handlerHealth)
		http.HandleFunc("/health.json", handlerHealth)

		/* API Version 1 */
		http.HandleFunc("/api", handlerAPIDefault)
		http.HandleFunc("/api/v1/health", handlerHealth)
		http.HandleFunc("/api/v1/nodes", handlerNodes)
		//http.HandleFunc("/api/v1/$node/measurements", handlerMeasurements)

		// Listen on http server
		go func() {
			log.Fatal(http.ListenAndServe(config.HttpBindAddr, nil))
		}()

		if config.Verbose {
			fmt.Printf("webserver enabled on http://%s\n", config.HttpBindAddr)
		}
	}

	fmt.Println("meteod is running")
	awaitTerminationSignal()
}

func replyJson(w http.ResponseWriter, obj interface{}) {
	buf, err := json.Marshal(obj)
	if err != nil {
		fmt.Fprintf(os.Stderr, "json marshall error: %s\n", err)
		w.WriteHeader(500)
		w.Write([]byte("server error"))
		return
	}
	w.Header().Add("Content-Type", "application/json")
	w.Header().Add("Content-Length", fmt.Sprintf("%d", len(buf)))
	w.WriteHeader(200)
	w.Write(buf)
}

func handlerDefault(w http.ResponseWriter, r *http.Request) {
	w.Write([]byte("meteod"))
}

func handlerAPIDefault(w http.ResponseWriter, r *http.Request) {
	content := []byte("<!DOCTYPE html>\n<html xmlns=\"http://www.w3.org/1999/xhtml\">\n<head><title>meteod api</title></head>\n<body>\n<h3>meteod api</h3>\n<p>Available routes</p>\n<ul>\n<li><a href=\"api/v1/health\">/api/v1/health</a></li>\n<li><a href=\"api/v1/nodes\">/api/v1/nodes</a></li>\n<li><a href=\"api/v1/measurements\">/api/v1/measurements</a></li></ul>\n</body>\n</html>")

	w.Header().Add("Content-Type", "text/html")
	w.Header().Add("Content-Length", fmt.Sprintf("%d", len(content)))
	w.Write(content)
}

func handlerNodes(w http.ResponseWriter, r *http.Request) {
	// Post array of current nodes without measurements
	type Node struct {
		ID   int    `json:"id"`
		Name string `json:"name"`
	}
	ret := make([]Node, 0)
	for _, m := range nodes {
		ret = append(ret, Node{ID: m.ID, Name: m.NodeName})
	}
	replyJson(w, ret)
}

func handlerHealth(w http.ResponseWriter, r *http.Request) {
	fmt.Fprintf(w, "{\"status\":\"ok\"}")
}
